/*
 * Crystal Mud Client
 * Copyright (C) 2002 Abigail Brady
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 */

/*
 *
 *   stdin -> utf-8 filter -> esc seq interpreter -> cmd interpreter
 *   read(telnet) -> telnet decoder -> utf-8 filter -> tty handler
 */

#define MCCP

#ifndef _XOPEN_SOURCE
#define _XOPEN_SOURCE
#endif

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <locale.h>
#include <wchar.h>
#include <langinfo.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <unistd.h>
#include <termios.h>
#include <signal.h>
#include <errno.h>
#include <netdb.h>

#include <netinet/in.h>

#define TELCMDS
#define TELOPTS
#include <arpa/telnet.h>
#undef TELCMDS
#undef TELOPTS

#undef WILL_LINEMODE

#include <arpa/inet.h>

#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>

#include <iconv.h>
#include <term.h>

#include <list>
#include <vector>
#include <map>
#include <string>
#include <set>

#include <stdarg.h>
#include <ctype.h>

#include <libintl.h>

#ifndef _
#define _(a) gettext(a)
#endif

using namespace std;

#include "Socket.h"

#include "config.h"

#undef SCROLL

FILE *logfile = 0;

#define debug_fprintf(a) ;

int info_to_stderr = 1;

void connect(const char *host, int port);
int file_dump(const char * file);
string mud_cset = "ISO-8859-1";
int allstars = 0;
int will_echo = 0;
int charmode = 0;
int rcvd_iac = 0;

#ifdef MCCP
#include "mccpDecompress.h"
static mc_state *mc = 0;
#endif

inline int MIN(int a, int b) {
  if (a>b) return b;
  return a;
}

#define MAXHEIGHT 200
#define MAXWIDTH 320

struct mterm {
  int utf8;
  int xterm_title;
  int col256;
  int knowscroll;
  string curtitle;

  mterm() : utf8(0), xterm_title(0), col256(0), knowscroll(0), curtitle(""),
  acsc(""), acsc_set(0) {
    HEIGHT = 24;
    WIDTH = 80;
  }

  void beep() {
    printf("\a");
    fflush(stdout);
  }

  void title(const char *fmt, ...) {
    if (xterm_title) {
      char buf[1000];
      va_list a;
      va_start(a, fmt);
      vsprintf(buf, fmt, a);
      if (curtitle != buf) {
	printf("\033]0;%s\a", buf);
	curtitle = buf;
      }
      va_end(a);
    }
  }
  
  string getinfo(char *name, const char *def=0) {
    const char *i = tigetstr(name);
    if (strstr(i, "$<")) {
      string q = i;
      q = q.substr(0, strstr(i, "$<")-i);
      return q;
    }
    if (i) 
      return i;
    return def;
  }

  string acsc;
  int acsc_set;

  void outvtchar(unsigned char w) {
    if (!acsc_set) {
      acsc = getinfo("acsc", "");
      acsc_set = 1;
    }
    const char *s = acsc.c_str();
    while (*s) {
      if (*s==w) {
	printf("%c", s[1]);
	return;
      }
      s += 2;
    }

    const char *def = "+#????\\+?#+++++~--_++++++<>*!Lo";
    if (w >= '`' && w <= 0x7e) {
      printf("%c", def[w-'`']);
      return;
    }

    printf("%c", '?');
  }

  int HEIGHT;
  int WIDTH;

  void grabwinsize() {
    struct winsize wws;
    const struct winsize *ws = &wws;
    ioctl(0, TIOCGWINSZ, ws);
    if (ws->ws_row)     HEIGHT = MIN(ws->ws_row,     MAXHEIGHT)-1;
    if (ws->ws_col)     WIDTH = MIN(ws->ws_col, MAXWIDTH);    
  }

  void getterm() {
    const char *t=getenv("TERM");
    if (t && (strcmp(t, "xterm")==0||
	      strcmp(t, "rxvt")==0||
	      strcmp(t, "xterm-color")==0||
	      strcmp(t, "screen")==0)||
	strncmp(t, "xterm-", 5)==0) {
      xterm_title = 1;
      knowscroll = 0;
      if (strstr(t, "256"))
	col256 = 1;
    }
  }
} tty;

typedef basic_string<wchar_t> my_wstring;

enum Int {
  I_DIM,
  I_NORM,
  I_BOLD,
};

struct cell_t { 
  wchar_t ch;
  enum Int inten;
  unsigned char fc;
  bool scs:1;
  unsigned char bc;
  bool ul:1;
  bool it:1;
  bool fr:1;
  bool os:1;
  bool inv:1;
  void operator= (const cell_t &o) {
    ch = o.ch;
    inten = o.inten;
    fc = o.fc;
    bc = o.bc;
    scs = o.scs;
    ul = o.ul;
    it = o.it;
    fr = o.fr;
    os = o.os;
    inv = o.inv;
  }
  bool operator==(const cell_t &o) const {
    if (ch != o.ch) return false;
    if (inten != o.inten) return false;
    if (fc != o.fc) return false;
    if (bc != o.bc) return false;
    if (scs != o.scs) return false;
    if (ul != o.ul) return false;
    if (it != o.it) return false;
    if (fr != o.fr) return false;
    if (inv != o.inv) return false;
    return true;
  }
  bool operator != (const cell_t &o) const {
    return !operator==(o);
  }
};


#define COL_BLACK   0
#define COL_RED     1
#define COL_GREEN   2
#define COL_YELLOW  3
#define COL_BLUE    4
#define COL_MAGENTA 5
#define COL_CYAN    6
#define COL_WHITE   7
#define COL_DEFAULT 9

struct cell_t blank = {' ', I_NORM, COL_DEFAULT, 0, COL_DEFAULT, 0 };
struct cell_t blank2 = {0, I_NORM, COL_DEFAULT, 0, COL_DEFAULT, 0 };

class cellstring {
  int els;
  int siz;
  cell_t *s;
public:
  cellstring() {
    els = 0;
    s = 0;
    siz = 0;
  }
  int length() { return els; }
  void operator+=(const cell_t &c) {
    if ((els+1)>siz) {
      siz *= 2;
      if (siz == 0) 
	siz = 16;
      s = (cell_t*)realloc(s, sizeof(cell_t)*siz);
    }
    s[els] = c;
    els++;
  }
  const cell_t &operator[](size_t i) {
    if (i >= els) return blank;
    return s[i];
  }
  ~cellstring() {
    if (s) free(s);
  }
};


int will_eor = 0;

#define GRIDHEIGHT 10000

int grid_changed = 0;
int buffer_changed = 0;

struct line_t {
  cell_t i[MAXWIDTH];
  int len;
};

cell_t mkchar(wchar_t ch, Int inten, int fc, int bc, int scs, int ul, int it, int fr, int os, int in)
{
  cell_t i = { ch, inten, fc, scs, bc, ul, it, fr, os, in };
  return i;
}

cell_t mkchar(wchar_t ch)
{
  cell_t i = { ch, I_NORM, COL_DEFAULT, 0, COL_DEFAULT, 0, 0, 0, 0, 0 };
  return i;
}

#undef lines
#undef newline
#undef grid

struct grid_t {
  vector<line_t> lines;
  int first;
  int n;
  int lcount;

  int row;
  int col;

  int lastprompt;

  grid_t() {
    first = 0;
    n = 0;
    lcount = GRIDHEIGHT;
    lines.reserve( lcount );
    lastprompt = 0;
    nlw = 0;
  }

  cell_t get(int r, int c) {
    if (r < first)
      return blank;
    if (r > (first + n))
      return blank;

    if (c > lines[r % lcount].len)
      return blank;

    return lines[r % lcount].i[c];
  }

  void newline() {
    if (logfile) {
      for (int i=0;i<get_len(row);i++) {
	fprintf(logfile, "%lc", get(row,i).ch);
      }
      fprintf(logfile, "\n");
      fflush(logfile);
    }

    row++;
    col = 0;
    n++;
    for (int i=0;i<MAXWIDTH;i++) {
      lines [ row % lcount ].i[i] = blank;
    }
    if (n > lcount) {
      first++;
      n--;
    }
  }

  void eraseline(int from) 
  {
    for (int i=from;i<MAXWIDTH;i++) {
      lines [ row % lcount ].i[i] = blank;
    }
  }

  int get_len(int r) {
    if (r < first)
      return 0;
    if (r > (first + n))
      return 0;

    return lines[r % lcount].len;
  }

  void set(int r, int c, cell_t ch) {
    if (lines [ r % lcount].len <= c) 
      lines[r % lcount].len = c+1;
    lines[r % lcount].i[c] = ch;
  }

  int nlw;

  void wantnewline() {
    nlw++;
  }

  void place(const cell_t *ri) {
    cell_t fi = *ri;
    cell_t *i = &fi;

    for (int i=0;i<nlw;i++) {
      newline();
    }
    nlw = 0;
    
    if (i->ch == '\r') {
      return;
    }
    if (i->ch == '\t') {
      set(row, col, blank);
      col++;
      while ((col & 7) && col < tty.WIDTH) {
	set(row, col, blank);
	col++;
      }
      return;
    }
    if (col >= tty.WIDTH || i->ch == '\n') {
      newline();
      if (i->ch == '\n')
	return;
    }
    
    if (((i->ch < 0x20) || (i->ch >= 0x80 && i->ch <= 0x9f)))
      i->ch = 0x241b;
    
    set(row, col, *i);
    col++;
    if (wcwidth(i->ch)==2) {
      set(row, col, mkchar(-i->ch));
      col++;
    } 

    lastprompt = 1;
  }
} grid;

struct conn_t {
  grid_t *grid;

  my_wstring buffer;
  int c;

  my_wstring cutbuffer;

  void doup();
  void dodown();
  void doleft();
  void doright();
  void dostart();
  void doclearline();
  void doeol();
  void dorefresh();
  void dotranspose();
  void doinsertchar(wchar_t);
  void dobackspace();
  
  void dokillword();
  void dodelete();

  void docutfromhere();
  void docuttohere();
  void dopaste();

  void dopageup();
  void dopagedown();
  void dohome();
  void doend();
  void triggerfn(const char *which);
  void findnext();

  conn_t() {
    grid = &::grid;

    buffer = L"";
    c = 0;

    cutbuffer = L"";
  }
} conn;

int commandmode = 0;

int hardscroll = 0;

static int cfg = COL_DEFAULT;
static int cbg = COL_DEFAULT;

static int cint = 0;
static int ccs = 0;
static int cul = 0;
static int cit = 0;
static int cfr = 0;
static int cinv = 0;
static int cos = 0;

static int died = 0;

bool plonk(const cell_t &g) {
  if (g.scs != ccs) {
    if (g.scs)
      printf("%s", tty.getinfo("smacs").c_str()); // setmode alternate character set
    else
      printf("%s", tty.getinfo("rmacs").c_str()); // resetmode alternate character set

    ccs = g.scs;
  }

  string whats = "";

  if ((g.fc != cfg && g.fc == COL_DEFAULT) ||
      (g.bc != cbg && g.bc == COL_DEFAULT)) {
    whats += "0;";
    cfg = COL_DEFAULT;
    cbg = COL_DEFAULT; 
    cint = 1;
    cul = 0;
    cit = 0;
    cfr = 0;
    cinv = 0;
    cos = 0;
  }

  if (g.inten != cint) {
    if (g.inten == I_BOLD)
      whats += "1;";
    else if (g.inten == I_DIM)
      whats += "2;";
    else
      whats += "22;";
    
    cint = g.inten;
  }

  if (g.fc != cfg) {
    static char blah[100];
    if (g.fc>7) {
      if (tty.col256)
	sprintf(blah, "31;");
      else
	sprintf(blah, "38;5;%i;", g.fc);
    } else {
      sprintf(blah, "%i;", 30+g.fc);
    }
    whats += blah;
    
    cfg = g.fc;
  }

  if (g.bc != cbg) {
    static char blah[100];
    if (g.bc>7) {
      if (tty.col256)
	sprintf(blah, "41;");
      else
	sprintf(blah, "48;5;%i;", g.bc);
    } else {
      sprintf(blah, "%i;", 40+g.bc);
    }
    whats += blah;
    
    cbg = g.bc;
  }

  if (g.ul != cul) {
    if (g.ul)
      whats += "4;";
    else
      whats += "24;";
    cul = g.ul;
  }

  if (g.it != cit) {
    if (g.it)
      whats += "3;";
    else
      whats += "23;";
    cit = g.it;
    cfr = 0;
  }

  if (g.inv != cinv) {
    if (g.inv)
      whats += "7;";
    else
      whats += "27;";
    cinv = g.inv;
  }

  if (g.os != cos) {
    if (g.os)
      whats += "9;";
    else
      whats += "29;";
    cos = g.os;
  }

  if (g.fr != cfr) {
    if (g.fr)
      whats += "20;";
    else
      whats += "23;";
    cfr = g.fr;
  }

  if (whats.length()) {
    whats = whats.substr(0, whats.length()-1);
    printf("\033[%sm", whats.c_str());
  }
  
  if (g.ch >= 32  && g.ch < 127) {
    if (g.scs) {
      tty.outvtchar(g.ch);
    } else
      printf("%c", g.ch);
  } else if(g.ch==0) {
    if (!died) {
      printf("%s", tty.getinfo("el").c_str()); // clear to end of line
      died = 1;
    }
  } else if (g.ch < 256 || tty.utf8) {
    wchar_t ch = g.ch;
    if (ch < 32)
      ch += 0x2400;
    printf("%lc", ch);
  } else {
    switch (wcwidth(g.ch)) {
    case 0:
      break;
    case 1:
      printf("-");
      break;
    case 2:
      printf("--");
      break;
    }
  }
  return 0;
}

cell_t wantbuffer[MAXHEIGHT][MAXWIDTH];
cell_t havebuffer[MAXHEIGHT][MAXWIDTH];

int evillines[MAXHEIGHT] = { 0,  };

int bad_have = 1;

void show_want() {
  
  int realy = -1;
  int realx = -1;

  if (tty.knowscroll) {
    int anyevil = 0;
    for (int i=0;i<tty.HEIGHT;i++) {
      int notsame = 0, onebelow = 1;
      for (int j=0;j<tty.WIDTH;j++) {
	if (wantbuffer[i][j]!=havebuffer[i+1][j]) {
	  onebelow = 0;
	}
	if (wantbuffer[i][j] != havebuffer[i][j]) {
	  notsame = 1;
	}
      }
      if (notsame && onebelow) {
	for (int j=0;j<tty.WIDTH;j++) {
	  havebuffer[i][j] = wantbuffer[i][j];
	  havebuffer[i+1][j] = blank;
	}
	evillines[i] = 0;
	evillines[i+1] = 1;
	printf("\033[40m");
	printf("\033[%i;%ir", i+1, i+2);
	printf("\033[1S");
	printf("\033[r");
	continue;
      }      
    }
  }
  
  cfg = -1;
  cbg = -1;
  cint = -1;
  ccs = -1;
  cul = -1;
  cit = -1;
  cfr = -1;
  cinv = -1;
  cos = -1;

  for (int i=0;i<(tty.HEIGHT+1);i++) {
    int wanty = i+1;
    int wantx = 1;
    died = 0;
    int blah = (MAXHEIGHT-tty.HEIGHT)-1;

    for (int j=0;j<((i==tty.HEIGHT)?tty.WIDTH-1:tty.WIDTH);j++) {
      if (havebuffer[i][j] == wantbuffer[i][j] && !bad_have
      	  && !evillines[i]) {
      	wantx++;
	continue;
      } else {
	if (realy != wanty) {
	  realy = wanty;
	  realx = wantx;
	  printf("\033[%i;%if", wanty, wantx);
	} else if (realx != wantx) {
	  if (wantx==(realx+1)) {
	    printf("\033[C");
	  } else {
	    printf("\033[%i;%if", wanty, wantx);
	  }
	  realx = wantx;
	}
	
	havebuffer[i][j] = wantbuffer[i][j];
	wantx++;
	realx++;
	plonk(wantbuffer[i][j]);
      }
    }
    evillines[i] = 0;
  } 
  bad_have = 0;
}

set<my_wstring> hl_matches;

void show_lines_at(int from, int to, int num)
{
  for (int i=0;i<num;i++) {
    if (i+from == grid.row && !charmode) {
      for (int j=0;j<tty.WIDTH;j++)
	wantbuffer[i+to-1][j] = blank2;
      continue;
    }
    int mw = MIN(tty.WIDTH, grid.get_len(i+from));
    int j;
    if (tty.utf8) mw++;
    for (j=0;j<mw;j++) {
      cell_t g = grid.get(i+from, j);
      wantbuffer[i+to-1][j] = g;
    }

    my_wstring s;
    for (int j=0;j<grid.get_len(i+from);j++) {
      s += grid.get(i+from, j).ch;
    }

    set<my_wstring>::iterator it;
    for (it=hl_matches.begin();it!=hl_matches.end();it++) {
      size_t l = s.find(*it);
      while (l != -1) {
	for (int j=0;j<it->length();j++)
	wantbuffer[i+to-1][j+l].inv = !wantbuffer[i+to-1][j+l].inv;
	l = s.find(*it, l+1);
      }
    }
      
    while (j<tty.WIDTH) {
      wantbuffer[i+to-1][j] = blank2;
      j++;
    }
  }
}

void conn_t::findnext()
{
  for (int i=hardscroll;i<grid->row;i++) {
    my_wstring s;
    for (int j=0;j<grid->get_len(i);j++) {
      s += grid->get(i, j).ch;
    }
    
    set<my_wstring>::iterator it;
    for (it=hl_matches.begin();it!=hl_matches.end();it++) {
      size_t l = s.find(*it);
      while (l != -1) {
	for (int j=0;j<it->length();j++) {
	  hardscroll = i+1;
	  if (hardscroll>(grid->row-tty.HEIGHT))
	    hardscroll = 0;
	  grid_changed = 1;
	  return;
	}
	l = s.find(*it, l+1);
      }
    }
  }
  hardscroll = 0;
  grid_changed = 1;
}

void display_buffer(conn_t *conn) {
  if (info_to_stderr) return;

  if (grid_changed) buffer_changed = 1;

  int curpos_changed = grid_changed | buffer_changed;

  const wchar_t *a = conn->buffer.data();
  int alen = conn->buffer.length();

  const int col = conn->c;
  const wchar_t *txt = a;

  int scroll = 0;

  int start = grid.row - tty.HEIGHT;
  if (start < 0) {
    start = 0;
  }

  if (hardscroll)
    start = hardscroll-1;

  if (grid_changed) {
    if (hardscroll) {
      show_lines_at(start, 1, (tty.HEIGHT-4));
      show_lines_at((grid.row-5)>0?grid.row-5:0, tty.HEIGHT-4, 5);
      for (int i=0;i<tty.WIDTH;i++) {
	wantbuffer[tty.HEIGHT-5][i] = mkchar('=', I_BOLD, COL_WHITE, COL_RED, 0, 0, 0, 0, 0, 0);
      }

    } else {
      if (grid.row<tty.HEIGHT)
	show_lines_at(start, tty.HEIGHT-grid.row+1, tty.HEIGHT);
      else
	show_lines_at(start, 1, tty.HEIGHT);
    }
    grid_changed = 0;
  }

  cellstring crealprompt;

  if (commandmode) {
    const char *str = _("crystal> ");
    int max = strlen(str);
    while (1) {
      wchar_t c;
      int s=mbtowc(&c, str, max);
      if (s<=0)
	break;
      crealprompt += mkchar(c);
      str += s;
      max -= s;
    }
  } else {
    if (!will_eor || grid.lastprompt) {
      for (int i=0;i<grid.col;i++) {
	crealprompt += grid.get(grid.row, i);
      }
    }
  }

  int prlen = crealprompt.length();
  
  if (prlen>(tty.WIDTH-20))
    prlen = (tty.WIDTH-20);

  int wid = tty.WIDTH-5;
  if (conn->c+prlen>=wid) {
    scroll = conn->c-wid+1+prlen;
    a += scroll;
    alen -= scroll;
  }
  wid =tty.WIDTH-1;
  
  if (buffer_changed) {
    int i;
    int pl = crealprompt.length();
    if (pl > (tty.WIDTH-20))
      pl =tty.WIDTH-20;
    for (i=0;i<pl;i++)
      {
	wantbuffer[tty.HEIGHT][i] = crealprompt[i];
      }
    
    wid -= prlen;
    if (!allstars || commandmode) {
      while (alen && wid) {
	if (wcwidth(*a)<=wid) {
	  wantbuffer[tty.HEIGHT][i++] = mkchar(*a);
	  if (wcwidth(*a)==2)
	    wantbuffer[tty.HEIGHT][i++] = mkchar(-*a);
	} else
	  break;
	wid -= wcwidth(*a);
	a++;
	alen--;
      }
    }
    while (wid) {
      wantbuffer[tty.HEIGHT][i++] = blank;
      wid--;
    }
    if (alen) {
      wantbuffer[tty.HEIGHT][i++] = mkchar('>');
    } else
      wantbuffer[tty.HEIGHT][i++] = mkchar(' ');
  }

  show_want();

  int rcol=col;

  for (int i=0;i<col;i++) {
    if (wcwidth(txt[i])==2)
      rcol++;
  }

  if (allstars && !commandmode) 
    rcol = 0;

  printf("\033[%i;%if", tty.HEIGHT+1, rcol-scroll+1+prlen);
  buffer_changed = 0;

  fflush(stdout);
}

#define BEL '\a'
#define SI  0x0e
#define SO  0x0f

#define LF  '\n'
#define VT  '\v'
#define FF  '\f'
#define CR  '\r'
#define BS  '\b'

#define ESC L'\033'
#define CSI L'\233'

enum {
  TITLEBAR = 256,
  TITLEBAR2
 
};

void info(const char *);
void infof(const char *fmt, ...) {
  char buf[10000];
  va_list ap;
  va_start(ap, fmt);
  vsprintf(buf, fmt, ap);
  va_end(ap);
  info(buf);
}
void info(const wchar_t *);

void show_batch(cellstring &batch) {
  for (int i=0;i<batch.length();i++) {
    grid.place(&batch[i]);
  }
}

void wterminal(wchar_t ch) {

  static int mode = 0;
  static list<int> pars;
  static int par;

  static Int inten = I_NORM;
  static int forecol = COL_DEFAULT;
  static int backcol = COL_DEFAULT;
  static int scs = 0;
  static int ul = 0;
  static int it = 0;
  static int fr = 0;
  static int inv = 0;
  static int os = 0;
  static int hidden = 0;
  static int lastcr = 0;

  static string title="";

  if (mode==0) {
    switch (ch) {
    case 0:
      return;

    case CR:
      grid.col = 0;
      return;

    case BS:
      grid.col--;
      if (grid.col < 0) grid.col = 0;
      return;

    case BEL:
      tty.beep();
      return;

    case SI:
      scs = 1;
      return;

    case SO:
      scs = 0;
      return;

    case ESC:
      mode = ESC;
      par = -1;
      {
      list<int> q;
      pars = q;
      }
      return;

    case CSI:
      mode = CSI;
      par = -1;
      {
      list<int> q;
      pars = q;
      }
      return;

    default:
      
      if (ch >= 0x80 && ch <= 0x9f)
	ch = '?';
      
      if (ch == LF || ch == FF || ch == VT) {
	ch = LF;
      }
      
      if (!will_eor)
	buffer_changed = 1;
      
      if (hidden) ch = ' ';
	
      cell_t a = mkchar(ch, inten, forecol, backcol, scs, ul, it, fr, os, inv);
      grid.place(&a);
      
      grid_changed = 1;
      if (!will_eor)
	buffer_changed = 1;
    }
    return;
  }
  
  if (mode == ESC) {
    if (ch == '[') {
      mode = CSI;
      return;
    }
    if (ch == ']') {
      mode = ']';
      return;
    }
    if (ch == '(') {
      mode = '(';
      return;
    }
    if (ch == '%') {
      mode = '%';
      return;
    }

    if (ch >= 48 && ch <= 126) {
      mode = 0;
      return;
    }
  }

  if (mode == '%') {
    if (ch == '@')
      mud_cset = "ISO-8859-1";
    
    if (ch == 'G')
      mud_cset = "UTF-8";

    mode = 0;

    return;
  }

  if (mode == '(') {
    mode = 0;
    if (ch == 'B') {
      scs = 0;
    }
    if (ch == '0') {
      scs = 1;
    }  
    return;
  }

  if (mode == ']') {
    if (ch != '0')
      mode = 0;
    else
      mode = TITLEBAR;
    return;
  }

  if (mode == TITLEBAR) {
    if (ch != ';')
      mode = 0;
    else {
      mode = TITLEBAR2;
      title = "";
    }
    return;
  }

  if (mode == TITLEBAR2) {
    if (ch != '\a') {
      title += ch;
    } else {
      tty.title(_("%s - Crystal"), title.c_str());
      mode = 0;
    }
    return;
  }

  if (mode == CSI) {

    if (isdigit(ch)) {
      if (par == -1) 
	par = (ch - '0');
      else
	par = (par * 10) + (ch - '0');
      return;
    }

    if (ch == ';') {
      pars.insert(pars.end(), par);
      par = -1;
      return;
    }    

    if (par!=-1)
      pars.insert(pars.end(), par);

    if (ch=='A') { /* CUU - CUrsor Up */
      grid.row--;
      mode = 0;
      return;
    }

    if (ch=='K') {  /* EL - Erase in Line */
      if (pars.begin() == pars.end())
	pars.insert(pars.end(), 0);
      
      /* 0 - default - cursor to end of line */
      /* 1 -           start to cursor (including) */
      /* 2 -           all of line */

      par = *pars.begin();
      if (par == 0)
	grid.eraseline(grid.col);
      else if (par == 2)
	grid.eraseline(0);
    }

    if (ch=='m') { /* SGR - Select Graphics Rendition */
      if (pars.begin() == pars.end()) {
	pars.insert(pars.end(), 0);
      }

      list<int>::iterator parit = pars.begin();
      while (parit != pars.end()) {
	int parm = *parit;
	parit++;
#define DEF_FC COL_WHITE
#define DEF_BC COL_DEFAULT
	switch (parm) {
	case 0:
	  inten = I_NORM;
	  forecol = DEF_FC;
	  backcol = DEF_BC;
	  ul = 0;
	  it = 0;
	  fr = 0;
	  inv = 0;
	  hidden = 0;
	  os = 0;
	  break;
	case 1:
	  inten = I_BOLD;
	  break;
	case 2:
	  inten = I_DIM;
	  break;
	case 3:
	  it = 1;
	  fr = 0;
	  break;
	case 4:
	  ul = 1;
	  break;
	case 7:
	  inv = 1;
	  break;
	case 8:
	  hidden = 1;
	  break;
	case 9:
	  os = 1;
	  break;
	case 20:
	  fr = 1;
	  it = 0;
	  break;
	case 22:
	  inten = I_NORM;
	  break;
	case 23:
	  fr = it = 0;
	  break;
	case 24:
	  ul = 0;
	  break;
	case 27:
	  inv = 0;
	  break;
	case 28:
	  hidden = 0;
	  break;
	case 29:
	  os = 0;
	  break;
	case 30 ... 37:
	  forecol = parm - 30;
	  break;
	case 39:
	  forecol = DEF_FC;
	  break;
	case 38:
	  if (parit != pars.end()) {
	    if (*parit == 5) {
	      parit++;
	      if (parit != pars.end()) {
		int c = *parit;
		if (c >= 0 && c <= 7)
		  forecol = c;
		else if (c >= 8 && c <= 15)
		  forecol = c&7;
		else {
		  forecol = c;
		}
		parit++;
		continue;
	      }
	    }
	  }
	  break;
	case 40 ... 47:
	  backcol = parm - 40;
	  break;
	case 49:
	  backcol = DEF_BC;
	  break;
	}
      }

      mode = 0;
      par = -1;
      list<int> q; pars = q;
      return;
    }

    if (ch>=64 && ch <= 126) {
      mode = 0;
      par = -1;
      list<int> q; pars = q;
    }

    return;
  }
}

void infoc(wchar_t w) {
  
  if (info_to_stderr) {
    fprintf(stderr, "%lc", w);
  }

  static my_wstring isf = L"";

  isf += w;

  if (w=='\n') {
    cellstring q;
    int c = -1;
    if (grid.col) {
      for (int i=0;i<grid.get_len(grid.row);i++) {
	q += grid.get(grid.row, i);
      }
      c = grid.col;
      grid.col = 0;
      grid.eraseline(0);
    }
    for (int i=0;i<isf.length();i++) {
      cell_t c = mkchar(isf[i]);
      grid.place(&c);
    }
    
    if (c != -1) {
      for (int i=0;i<q.length();i++)
	grid.place(&q[i]);
      c = grid.col;
    }

    grid_changed = 1;
    isf = L"";
    return;
  }
}

void info(const char *str) {  
  int max = strlen(str);
  while (1) {
    wchar_t c;
    int s=mbtowc(&c, str, max);
    if (s<=0)
      return;
    infoc(c);
    str += s;
    max -= s;
  }
  
}

void info(const wchar_t *w)
{
  while (*w) {
    infoc(*w);
    w++;
  }
}

void info(const my_wstring &str) {  
  for (int i=0;i<str.length();i++) {
    infoc(str[i]);
  }
}

void wtstring(const my_wstring &str) {  
  int len = str.length();
  int i = 0;
  while (len) {
    wterminal(str[i]);
    len--;
    i++;
  }
}


void have_prompt() {
  grid.lastprompt = 1;
  buffer_changed = 1;
}

Socket *s;

string nam(int i);

void reply(int a, int b, int c) {
  char buf[5];
  buf[0] = a;
  buf[1] = b;
  buf[2] = c;
  s->write(buf, 3);
  debug_fprintf((stderr, "SEND %s %s %s\n", nam(a).c_str(), nam(b).c_str(), nam(c).c_str()));
}

int DO_NAWS = 0;

void sendwinsize() {
  if (!DO_NAWS) 
    return;

  if (!s)
    return;
  
  char buf[20];
  buf[0] = IAC;
  buf[1] = SB;
  buf[2] = TELOPT_NAWS;

  int l = 3;

  char q[4];
  q[0] = (tty.WIDTH >> 8);
  q[1] = (tty.WIDTH & 0xff);
  q[2] = (tty.HEIGHT+1) >> 8;
  q[3] = (tty.HEIGHT+1) & 0xff;

  for (int i=0;i<4;i++) {
    unsigned char c = q[i];
    buf[l++] = c;
    if (c==IAC) buf[l++] = c;
  }

  buf[l++] = IAC;
  buf[l++] = SE;

  s->write(buf, l);
}

void sendttype() {
  char buf[30];
  sprintf(buf, "%c%c%c%c%s%s%c%c", IAC, SB, TELOPT_TTYPE, TELQUAL_IS,
	  mud_cset=="UTF-8"?"ucryotel":"cryotel", 
	  tty.col256?":256c":"",
	  IAC, SE);
  s->write(buf, strlen(buf+4)+4);
}

void decode(int ch) {
  if (ch<=0x7f || mud_cset=="ISO-8859-1") {
    wterminal(ch);
    return;
  }

  static iconv_t i=0;
  static string cur_cset = "";

  static char ibuf[10] = {0};
  char *in=ibuf;
  static size_t inb = 0;

  if (!i || cur_cset != mud_cset) {
    if (i)
      iconv_close(i);
    i = iconv_open("WCHAR_T", mud_cset.c_str());
    cur_cset = mud_cset;
    ibuf[0] = 0;
    inb = 0;
  }

  if (!i) {
    wterminal('?');
  }
    
  ibuf[inb++] = ch;
  if (inb>9) inb = 9;
  
  char obuf[5];
  char *out=obuf;
  size_t outb = 4;
  
  if (iconv(i, &in, &inb, &out, &outb)==-1) {
    if (errno==EILSEQ) {
      for (int id=0;id<inb;id++) {
	wterminal(0xfffd);
      }
      inb = 0;
      iconv(i, NULL, NULL, NULL, NULL);
    }
  }
  
  if (outb == 0) {
    wterminal(*((wchar_t*)obuf));
  }
  return;
}

string nam(int s) {
  if (s > xEOF)
    return telcmds[s - xEOF];
  if (s < NTELOPTS)
    return telopts[s];
  char q[4];
  sprintf(q, "%02x", s);
  return q;
}

#define MXP 0x5b

void tstack(int ch) {
  
  static int mode = 0;
  static int subneg_type = 0;
  static string subneg_data = "";

  if (mode == IAC) {
    debug_fprintf((stderr, "%s ", nam(ch).c_str()));

    if (ch == EOR || ch == GA) {
      will_eor = 1;
      have_prompt();
    }
    if (ch == NOP) {
      mode = 0;
      debug_fprintf((stderr, "\n"));
      return;
    }
    if (ch == IAC) {
      if (subneg_type)
	subneg_data += ch;
      else
	decode(0xff);
      mode = 0;
      debug_fprintf((stderr, "\n"));
      return;
    }
    if (ch == WILL) {
      mode = WILL;
      return;
    }
    if (ch == WONT) {
      mode = WONT;
      return;
    }
    if (ch == DO) {
      mode = DO;
      return;
    }
    if (ch == DONT) {
      mode = DONT;
      return;
    }
    if (ch == SB) {
      mode = SB;
      return;
    }
    if (ch == SE) {
      if (subneg_type = TELOPT_TTYPE) {
	sendttype();
      }
      subneg_type = 0;
      mode = 0;
      debug_fprintf((stderr, "\n"));
      return;
    }
    mode = 0;
    debug_fprintf((stderr, "\n"));
    return;
  }

  if (mode == SB) {
    debug_fprintf((stderr, "%s ", nam(ch).c_str()));
    subneg_type = ch;
    mode = 0;
    debug_fprintf((stderr, "\n"));
    return;
  }

  if (mode == WILL) {
    debug_fprintf((stderr, "%s ", nam(ch).c_str()));
    if (ch == TELOPT_ECHO) {
      allstars = 1;
      will_echo = 1;
      mode = 0;
      reply(IAC, DO, TELOPT_ECHO);
      debug_fprintf((stderr, "\n"));
      return;
    }

#ifdef WILL_LINEMODE
    if (!rcvd_iac)
      reply(IAC, WILL, TELOPT_LINEMODE);
#endif
    rcvd_iac = 1;

    if (ch == TELOPT_EOR) {
      debug_fprintf((stderr, "\n"));
      if (!will_eor) {
	reply ( IAC, DO, TELOPT_EOR );
	will_eor = 1;
      }
      mode = 0;
      return;
    }

#if 0
    if (ch == MXP) {
      reply( IAC, DO, MXP );
    }
#endif

    mode = 0;
    debug_fprintf((stderr, "\n"));
    return;
  }

  if (mode == WONT) {
    debug_fprintf((stderr, "%s ", nam(ch).c_str()));
    if (ch == TELOPT_ECHO) {
      allstars = 0;
      will_echo = 0;
      reply(IAC, DONT, TELOPT_ECHO);
    } else {
#ifdef WILL_LINEMODE
      if (!rcvd_iac)
	reply(IAC, WILL, TELOPT_LINEMODE);
#endif
      rcvd_iac = 1;
    }

    mode = 0;
    debug_fprintf((stderr, "\n"));
    return;
  }

  if (mode == DO) {
    debug_fprintf((stderr, "%s ", nam(ch).c_str()));

#ifdef WILL_LINEMODE
    if (!rcvd_iac)
      reply(IAC, WILL, TELOPT_LINEMODE);
#endif
    rcvd_iac = 1;

    if (ch == TELOPT_NAWS) {
      debug_fprintf((stderr, "\n"));
      
      reply( IAC, WILL, TELOPT_NAWS);

      DO_NAWS = 1;
      sendwinsize();
      mode = 0;
      return;
    }
    if (ch == TELOPT_TTYPE) {
      debug_fprintf((stderr, "\n"));
      static int ack2 = 0;
      if (!ack2) {
	reply ( IAC, WILL, TELOPT_TTYPE );
	ack2 = 1;
      }
      mode = 0;
      return;
    }

    debug_fprintf((stderr, "\n"));

    reply (IAC, WONT, ch);

    mode = 0;
    return;
  }

  if (mode == DONT) {
    debug_fprintf((stderr, "%s ", nam(ch).c_str()));
    debug_fprintf((stderr, "\n"));

#ifdef WILL_LINEMODE
    if (!rcvd_iac)
      reply(IAC, WILL, TELOPT_LINEMODE);
#endif
    rcvd_iac = 1;
#ifdef WILL_LINEMODE
    if (ch == TELOPT_LINEMODE)
      charmode = 1;
#endif
    mode = 0;
    return;
  }

  if (ch == IAC) {
    debug_fprintf((stderr, "RECV %s ", nam(ch).c_str()));
    mode = IAC;
    return;
  }

  if (!subneg_type) {
    decode(ch);
  } else {
    subneg_data += ch;
  }
}

void decompress(unsigned char *bytes, size_t len) {
#ifdef MCCP
  mudcompress_receive(mc, (const char *)bytes, len);
  while (mudcompress_pending(mc)) {
    char buf[2048];
    int got = mudcompress_get(mc, buf, 2048);
    for (int i=0;i<got;i++) {
      tstack((unsigned char)buf[i]);
    }
  }
  if (const char *st=mudcompress_response(mc)) {
    s->write(st, strlen(st));
  }
#else
  for (int i=0;i<len;i++)
    tstack(bytes[i]);
#endif
}


int getinput() {
  unsigned char buf[10];
  if (read(0, buf, 1)==1) {

    if (tty.utf8) {
      if (buf[0]<0x80)
	return buf[0];

      if (buf[0]>0xc0 && buf[0]<=0xdf) {
	read(0, buf+1, 1);
	buf[2] = 0;

	int ucs = (buf[0] - 0xc0) << 6;
	ucs |= (buf[1] & 63);
	return ucs;
      }
      if (buf[0]>=0xe0) {
	read(0, buf+1, 2);
	buf[3] = 0;

	int ucs = (buf[0] - 0xe0) << 12;
	ucs |= (buf[1] & 63) << 6;
	ucs |= (buf[2] & 63);

	return ucs;
      }
    }
    return buf[0]; 
  } else
   return -1;
}

void conn_t::triggerfn(const char *fn) {
  if (!s)
    return;
  string st = "fn.";
  st += fn;
  st += "\r\n";
  s->write(st.c_str(), st.length());
}


struct hlist {
  map<int, my_wstring> hist;
  int idx;
  int max;
  hlist() : idx(0), max(0) {
  }

  void insert(const my_wstring &blah) {
    hist[max] = blah;
    max++;
    idx = max;
  }

  bool back() {
    idx--;
    if (idx < 0) {
      idx = 0;
      return 0;
    }
    return 1;
  }

  bool next() {
    idx++;
    if (idx >= max) {
      idx = max;
      return  0;
    }
    return 1;
  }

  const my_wstring &get() {
    static const my_wstring blah = L"";
    if (idx >= max || idx < 0) 
      return blah;
    return hist[idx];
  }
};


hlist cmdhist;

hlist *chist() {
  static hlist hist;
  
  if (commandmode)
    return &cmdhist;

  return &hist;
}

string mks(const my_wstring &w) {
  string s;
  for (int i=0;i<w.length();i++) {
    wchar_t c = w[i];
    if (c < 0x80)
      s += c;
    else
      s += '?';
  }
  return s;
}

void conn_t::dokillword() {
  int orig=c;
  if (c) {
    while(c && isspace(buffer[c-1]))
      c--;
    while(c && !isspace(buffer[c-1]))
      c--;

    my_wstring extra = buffer.substr(orig);
    buffer = buffer.substr(0, c);
    buffer += extra;
  }
}

void conn_t::dodelete() {
  if (c < buffer.length()) {
    buffer.erase(c, 1);
  } else {
    tty.beep();
  }
}

void conn_t::dobackspace()
{
  if (c) {
    buffer.erase(c-1, 1);
    c--;
  } else {
    tty.beep();
  }
}

void conn_t::dohome() {
  if (grid->row < tty.HEIGHT)
    return;
  hardscroll = 1;
  grid_changed = 1;
}

void conn_t::doend() {
  hardscroll = 0;
  grid_changed = 1;
}

void conn_t::doup() {
  if (chist() && chist()->back()) {
    buffer = chist()->get();
    c = buffer.length();
  }
  else
    tty.beep();
}

void conn_t::dodown()
{
  if (chist() && chist()->next()) {
    buffer = chist()->get();
    c = buffer.length();
  }
  else
    tty.beep();
}

void conn_t::doleft()
{
  c--;
  if (c < 0) {
    c = 0;
    tty.beep();
  }
}

void conn_t::doright()
{
  c++;
  if (c>buffer.length()) {
    c = buffer.length();
    tty.beep();
  }
}

void conn_t::dostart() {
  c = 0;
}

void conn_t::doclearline() {
  buffer = L"";
  c = 0;
}

void conn_t::doeol() {
  c = buffer.length();
}

void conn_t::dorefresh() {
    grid_changed=1;
    buffer_changed=1;
    bad_have = 1;
}

void conn_t::dotranspose() {
  if (c>1 && buffer.length()==c) {
    wchar_t tmp = buffer[c-1];
    buffer[c-1] = buffer[c-2];
    buffer[c-2] = tmp;
    buffer_changed = 1;
  } else if (c>0) {
    wchar_t tmp = buffer[c];
    buffer[c] = buffer[c-1];
    buffer[c-1] = tmp;
    c++;
    buffer_changed = 1;
  } else {
    tty.beep();
  }
}

void conn_t::doinsertchar(wchar_t ch)
{
  if (c == buffer.length()) {
    buffer += ch;
  } else {
    buffer = buffer.substr(0, c) + ch + buffer.substr(c);
  }
  buffer_changed = 1;
  c++;
}

void cmd_quit(conn_t *conn, int &quit) {
  quit = 1;
}

void cmd_z(conn_t *conn, int &quit)
{
  kill(0, SIGTSTP);

  struct termios ti;
  cfmakeraw(&ti);
  tcsetattr(0, TCSADRAIN, &ti);
  
  tty.grabwinsize();

  bad_have = 1;
  buffer_changed = 1;
  grid_changed = 1;
  show_want();
}

void cmd_close(conn_t *conn, int &quit)
{
  if (!s) {
    info(_("/// you aren't connected.\n"));
    return;
  }

  delete s;
  s = 0;
  info(_("/// connection closed.\n"));
}

void cmd_help(conn_t *conn, int &quit) {
  info(_("/// connect <host> <port>  -- connect to a given host/port\n"
       "/// close                  -- close the current connection\n"
       "/// quit                   -- exit\n"
       "/// z                      -- suspend\n"
       "/// dump <file>            -- dump the backbuffer to a file. appends.\n"
       "/// log <file>             -- sets logfile as specified. appends.\n"
	 "/// charset <charset>      -- interprets data from the mud as given charset\n"));
#ifdef MCCP
  info(_("/// compress               -- show MCCP compression status \n"));
#endif
}

#ifdef MCCP
void cmd_compress(conn_t *conn, int &quit) {
  if (mc && mudcompress_compressing(mc)) {
    unsigned long compread, uncompread;
    mudcompress_stats(mc, &compread, &uncompread);
    infof(_("/// wire:%li bytes data:%li bytes.\n"), compread, uncompread);
  } else {
    info(_("/// no compression\n"));
  }
}
#endif

void cmd_connect(conn_t *conn, int &quit)
{
  my_wstring host = conn->buffer;
  host = host.substr(host.find(' ')+1);
  my_wstring port = host.substr(host.find(' ')+1);
  host = host.substr(0, host.find(' '));
  
  string cport = mks(port);
  string chost = mks(host);
  
  connect(chost.c_str(), atoi(cport.c_str()));
}

void cmd_match(conn_t *conn, int &quit) {
  my_wstring host = conn->buffer;
  if (host.find(' ')==-1)
    {
      set<my_wstring> nomatch;
      hl_matches = nomatch;
    }
  else {
    host = host.substr(host.find(' ')+1);
    if (hl_matches.find(host)==hl_matches.end())
      hl_matches.insert(host);
  }
  buffer_changed = 1;
  grid_changed = 1;
}

void cmd_charset(conn_t *conn, int &quit) 
{
  my_wstring filename = conn->buffer;
  filename = filename.substr(filename.find(' ')+1);
  
  string cfilename = mks(filename);
  mud_cset = cfilename;
  infof(_("/// charset '%s' selected\n"), mud_cset.c_str());
}  

void cmd_dump(conn_t *conn, int &quit)
{
  my_wstring filename = conn->buffer;
  filename = filename.substr(filename.find(' ')+1);
  
  string cfilename = mks(filename);
  file_dump(cfilename.c_str());
}

void cmd_log(conn_t *conn, int &quit)
{
  my_wstring filename = conn->buffer;
  filename = filename.substr(filename.find(' ')+1);
  
  string cfilename = mks(filename);
  //  file_dump(cfilename.c_str());
  if (logfile) {
    info(_("/// closing old logfile.\n"));
    fclose(logfile);
    logfile = NULL;
  }
  logfile = fopen(cfilename.c_str(), "a");
  if (!logfile) {
    infof(_("/// couldn't open '%s' for appending.\n"), cfilename.c_str());
  } else {
    infof(_("/// logging to end of '%s'.\n"), cfilename.c_str());
  }
}

struct {
  const wchar_t *commandname;
  void (*function)(conn_t *conn, int &quit);
} cmd_table[] =
{
#ifdef MCCP
  { L"compress", cmd_compress },
#endif
  { L"quit",     cmd_quit },
  { L"connect ", cmd_connect },
  { L"log ", cmd_log },
  { L"match", cmd_match },
  { L"close",    cmd_close },
  { L"open ",    cmd_connect },
  { L"dump ",    cmd_dump },
  { L"charset ", cmd_charset },
  { L"help",     cmd_help },
  { L"z",        cmd_z },
  { NULL, }
};

void doenter(conn_t *conn, int &quit)
{
  
  if (commandmode) {
    chist()->insert(conn->buffer);

    commandmode = 0;
    int handled = 0;
    int i = 0;
    info(_("crystal> "));
    info(conn->buffer);
    info("\n");

    const wchar_t *cmd = conn->buffer.data();
    int cmdlen = conn->buffer.length();

    wchar_t blah[cmdlen+2];
    wcsncpy(blah, conn->buffer.data(), cmdlen);
    blah[cmdlen] = 0;

    while (const wchar_t *cm=cmd_table[i].commandname) {
      if (wcsncmp(blah, cm, wcslen(cm))==0) {
	cmd_table[i].function(conn, quit);
	handled = 1;
	break;
      }
      i++;
    }
    if (!handled && conn->buffer.length()) {
      info(_("/// don't understand that command\n"));
    }
    conn->doclearline();
    return;
  }
  
  my_wstring wproper = L"";
  string proper = "";
  
  for (int i=0;i<conn->buffer.length();i++) {
    wchar_t sb = conn->buffer[i];

    wproper += sb;
    
    if (sb < 0x80) {
      proper += sb;
    }  else {

      static iconv_t i=0;
      static string cur_cset = "";
    
      if (!i || cur_cset != mud_cset) {
	if (i)
	  iconv_close(i);
	i = iconv_open(mud_cset.c_str(), "WCHAR_T");
	cur_cset = mud_cset;
      }
      wchar_t ic = sb;
      char *ibuf = (char*)&ic;
      size_t ilen = 4;
      char o[11];
      char *obuf = o;
      size_t olen = 10;
      iconv(i, &ibuf, &ilen, &obuf, &olen);
      if (olen != 10) {
	*obuf = 0;
	proper += o;
      }
    }
  }
  if (!will_echo) {
    
    if (wproper.length()) {
      chist()->insert(wproper);
    }

    my_wstring a;
    for (int i=0;i<grid.col;i++) {
      a += grid.get(grid.row, i).ch;
    }
    /* 
    if (a == L"[Press return to continue or 'q' to abort]" && wproper==L"") {
      grid.eraseline(0);
      grid.col = 0;
      grid.row-=2;
    }
    */
    
    for (int i=0;i<wproper.length();i++) {
      cell_t c = mkchar(wproper[i]);
      grid.place(&c);
    }
    grid.wantnewline();
    buffer_changed = 1;
    grid_changed = 1;
  }

  proper += "\r\n";
  if (s) {
    string p2;
    const unsigned char *b = (const unsigned char *)proper.c_str();
    while (*b) {
      if (*b==IAC) p2 += IAC;
      p2 += *b;
      b++;
    }
    s->write(p2.c_str(), p2.length());
  }

  conn->doclearline();
}

void conn_t::dopagedown()
{
  if (hardscroll) {
    hardscroll += 10;
  } 
  if (hardscroll > (grid->row - tty.HEIGHT))
    hardscroll = 0;	      
  grid_changed = 1;
}

void conn_t::dopageup()
{
  if (grid->row < tty.HEIGHT)
    return;

  if (hardscroll) {
    hardscroll -= 10;
  } else {
    hardscroll = (grid->row - tty.HEIGHT) - 9;
  }
  if (hardscroll < 1)
    hardscroll = 1;
  grid_changed = 1;
}

void conn_t::docutfromhere() {
  cutbuffer=buffer.substr(c);
  buffer = buffer.substr(0, c);
}

void conn_t::docuttohere() {
  cutbuffer = buffer.substr(0, c);
  buffer = buffer.substr(c);
  c = 0;
}

void conn_t::dopaste() 
{
  buffer.insert(c, cutbuffer);
}

void input(int i, conn_t *conn, int &quit) {
  int &c = conn->c;


  if (i=='')
    {
      if (!commandmode)
	conn->doclearline();
      commandmode = 1;
      return;
    }

  /* Code for character mode */
  if (!commandmode && charmode) {
    string p;
    p += i;
    s->write(p.c_str(), p.length());
    return;
  }

  switch (i) {

  case '\b': /* CONTROL H and DEL - backspace */
  case 127:
    conn->dobackspace();
    break;
    
  case 1: /* CONTROL A - to start of line */
    conn->dostart();
    break;

  case 2: /* CONTROL B - backward one char */
    conn->doleft();
    break;

  case 3: /* CONTROL C - clear line */
    conn->doclearline();
    break;

  case 4: /* CONTROL D - delete character */
    conn->dodelete();
    break;
	    
  case 5: /* CONTROL E - end of line */
    conn->doeol();
    break;

  case 6: /* CONTROL F - forward one char */
    conn->doright();
    break;

  case 9: /* CONTROL I - tab */
    conn->findnext();
    break;

  case 11: /* CONTROL K - cut from here */
    conn->docutfromhere();
    break;

  case 12: /* CONTROL L - refresh screen */
    conn->dorefresh();
    // also make it clear the screen?
    break;

  case 14: /* CONTROL N - next history item */
    conn->dodown();
    break;

  case 16: /* CONTROL P - previous history */
    conn->doup();
    break;

  case 20: /* CONTROL T - transpose */
    conn->dotranspose();
    break;

  case 21: /* CONTROL U - cut to here */
    conn->docuttohere();
    break;

  case 23: /* CONTROL W - delete word */
    conn->dokillword();
    break;
  
  case 25: /* CONTROL Y - paste */
    conn->dopaste();
    break;

  case 26: /* CONTROL Z - suspend */
    cmd_z(conn, quit);
    break;
    
  case '\n':
    break;
    
  case '\r':
    doenter(conn, quit);
    break;
    
  default:
    if (i < 32)
      break;

    conn->doinsertchar((wchar_t)i);
    break;
           
  case ESC:
    switch (getinput()) {

    case 'O':
      {
	int next = getinput();
	string proper = "";

	if (next == 'A') conn->doup();
	if (next == 'B') conn->dodown();

	if (next == 'C') conn->doright();
	if (next == 'D') conn->doleft();

	if (next == 'F') conn->doend();
	if (next == 'H') conn->dohome();
	if (next == 'P') conn->triggerfn("1");
	if (next == 'Q') conn->triggerfn("2");
	if (next == 'R') conn->triggerfn("3");
	if (next == 'S') conn->triggerfn("4");
	break;
      }

    case '[' :
      {
	int next = getinput();
	if (next == '[') {
	  next = getinput();
	  if (next == 'A') conn->triggerfn("1");
	  if (next == 'B') conn->triggerfn("2");
	  if (next == 'C') conn->triggerfn("3");
	  if (next == 'D') conn->triggerfn("4");
	  if (next == 'E') conn->triggerfn("5");
	  if (next == 'Z') /* conn->findprev() */;
	  return;
	}

	if (next == 'A')
	  conn->doup();

	if (next == 'B') 
	  conn->dodown();

	if (next == 'D')
	  conn->doleft();

	if (next == 'C')
	  conn->doright();

	string nums;

	if (isdigit(next)) {
	  while (isdigit(next)) {
	    nums += next;
	    next = getinput();
	  }
	  int num = atoi(nums.c_str());

	  if (next=='~') {
	    switch (num) {
	    case 3:
	      conn->dodelete();
	      break;
	    case 1:
	      conn->dohome();
	      break;
	    case 6:
	      conn->dopagedown();
	      break;
	    case 5:
	      conn->dopageup();
	      break;
	    case 4:
	      conn->doend();
	      break;

	    case 11: conn->triggerfn("1"); break;
	    case 12: conn->triggerfn("2"); break;
	    case 13: conn->triggerfn("3"); break;
	    case 14: conn->triggerfn("4"); break;
	    case 15: conn->triggerfn("5"); break;
	    case 17: conn->triggerfn("6"); break;
	    case 18: conn->triggerfn("7"); break;
	    case 19: conn->triggerfn("8"); break;
	    case 20: conn->triggerfn("9"); break;
	    case 21: conn->triggerfn("10"); break;
	    case 23: conn->triggerfn("11"); break;
	    case 24: conn->triggerfn("12"); break;
	    case 25: conn->triggerfn("13"); break;
	    case 26: conn->triggerfn("14"); break;
	    case 28: conn->triggerfn("15"); break;
	    case 29: conn->triggerfn("16"); break;
	    case 31: conn->triggerfn("17"); break;
	    case 32: conn->triggerfn("18"); break;
	    case 33: conn->triggerfn("19"); break;
	    case 34: conn->triggerfn("20"); break;
	    }
	  }
	}
	return;
      }
      break;
    }
  }  
}

void winch(int) {

  tty.grabwinsize();

  buffer_changed = 1;
  grid_changed = 1;
  bad_have = 1;

  sendwinsize();
}

void dispose() {
#ifdef MCCP
  if (mc) {
    mudcompress_delete(mc);
    mc = 0;
  }
#endif
  if (s) {
    delete s;
    s = 0;
  }
}

void connect(const char *host, int port)
{
  infof(_("/// resolving %s:%i\n"), host, port);
  display_buffer(&conn);
  
  dispose();

  InAddr *addr = InAddr::resolv(host);
  if (!addr) {
    infof(_("/// unknown host.\n"));
    return;
  }

  infof(_("/// connecting to %s:%i\n"), addr->tostring().c_str(), port);
  display_buffer(&conn);

  Socket *s2 = new Socket();
  addr->set_port(port);
  int stat = s2->connect(addr);
  delete addr;
  
  int e = errno;
  if (stat < 0) {
    infof(_("/// unable to connect : %s.\n"), strerror(e));
    
    delete s2;
    return;
  }

  s = s2;

  will_eor = 0;

  infof(_("/// connected\n"), host, port);

  static int a = 0;
  if (!a) {
    infof(_("/// escape character is '%s'\n"), "^]");
    a = 1;
  }

  tty.title(_("telnet://%s:%i - Crystal"), host, port);
#ifdef MCCP
  if (!mc) {
    mc = mudcompress_new();
  }
#endif

  rcvd_iac = 0;

  return;
}

int file_dump(const char * file){
  FILE *dumpfile; 
  
  dumpfile = fopen(file,"a");
  if(NULL==dumpfile){
    infof(_("/// couldn't open '%s' to dump to\n"), file);
    return 0;
  }
  infof(_("/// dumping scroll history to '%s'\n"), file);
  
  time_t cur_time = time(NULL);
  string ctime_str = ctime(&cur_time);
  ctime_str = ctime_str.substr(0,ctime_str.length()-1);
    
  fprintf(dumpfile,_("---=== Dump generated on %s by crystal ===---\n"),
	   ctime_str.c_str());
  for(int i = grid.first;i<grid.row;i++){
    for(int j=0;j<grid.get_len(i);j++)
      fprintf(dumpfile,"%lc",grid.get(i,j).ch);
    fprintf(dumpfile,"\n");
  }
  fprintf(dumpfile,_("---=== End of dump ===---\n"));
  fclose(dumpfile);
  return 1;
}

void main_loop() 
{
  if (!s) 
    commandmode = 1;

  buffer_changed = 1;
  display_buffer(&conn);
  
  int quit = 0;
  while (!quit) {
    fd_set r, e;

    int maxfd = 1;

    FD_ZERO(&r);
    FD_ZERO(&e);

    if (s) {
      FD_SET(s->getfd(),&r);
      FD_SET(s->getfd(),&e);

      maxfd = s->getfd() + 1;
    }

    FD_SET(0, &r);

    struct timeval tm;
    if (grid.nlw) {
      tm.tv_usec = 100000;
      tm.tv_sec = 0;
    } else {
      tm.tv_sec = 1;
      tm.tv_usec = 0;
    }
    select(maxfd, &r, 0, &e, &tm);
    
    for (int i=0;i<grid.nlw;i++) {
      grid.newline();
      grid_changed = 1;
    }
    grid.nlw = 0;

    if (s && (FD_ISSET(s->getfd(), &r) || FD_ISSET(s->getfd(), &e))) {
    int ok = 0;
    do {
	unsigned char mbuffer[1000];
	int bts = s->read((char *)mbuffer, 500);
	ok = 0;
	if (bts == 0 || (bts==-1 && s->getdead())) {
	  if (bts==-1) {
	    int e = errno;
	    infof(_("/// connection closed : %s.\n"), strerror(e));
	  } else {
	    info(_("/// connection closed by foreign host.\n"));
	  }
	  display_buffer(&conn);
	  fflush(stdout);
	  ok = 0;
	  dispose();
	  commandmode = 1;
	  buffer_changed = 1;
	  
	  conn.c = 0;
	  conn.buffer = L"";

	  commandmode = 1;
	  break;
	} else if (bts!=-1) {
	  decompress(mbuffer, bts);
	  ok = 1;
	}
      } while (ok);
    }

    if (FD_ISSET(0, &r)) {
      
      while (1) {
	int foo;
	ioctl(0, FIONREAD, &foo);
	if(foo) {
	  wchar_t i = getinput();

	  input(i, &conn, quit);

	  if (!s) {
	    commandmode = 1;
	  }
	  buffer_changed = 1;
	} else {
	  break;
	}
      }
    }
  display:
    if (!quit) {
      display_buffer(&conn);
      fflush(stdout);
    }
  }
}

int main(int argc, char **argv) {
  char *pname = argv[0];

  setlocale(LC_ALL, "");
  const char *codeset = nl_langinfo(CODESET);
  if (strcmp(codeset, "UTF-8")==0) {
    tty.utf8 = 1;
    mud_cset = "UTF-8";
  }

  tty.getterm();
#ifdef I18N
  bindtextdomain(PACKAGE, LOCALEDIR);
  textdomain(PACKAGE);
#endif
  if (argc>1) {

    if (strncmp(argv[1], "telnet://", 9)==0) {
      argv[1] += 9;
      if (char *s=strchr(argv[1], '/')) {
	*s = 0;
      }
    }
    
    char *pstr = argv[2];

    if (argc<=2)
      pstr = "telnet";

    if (strchr(argv[1], ':')) {
      pstr = strchr(argv[1], ':');
      *pstr = 0;
      pstr++;
    }

    char *e = 0;
    int port = strtol(pstr, &e, 10);
    if (*e) {
      struct servent *se = getservbyname(pstr, "tcp");
      if (se) {
	port = ntohs(se->s_port);
      } else {
	fprintf(stderr, _("%s: Bad port - '%s'.\n"), pname, pstr);
	exit(1);
      }
    }

    connect(argv[1], port);
    if (!s)
      exit(1);
#if 0
    if (er==1) {
      fprintf(stderr, _("%s: Unknown host.\n"), pname);
      exit(1);
    }
    
    if (er < 0) {
      fprintf(stderr, _("%s: Unable to connect to remote host: %s\n"), pname, strerror(-er));
      exit(1);
    }
#endif    
    wchar_t blah[1000];
    swprintf(blah, 1000, L"connect %s %s", argv[1], argv[2]);
    cmdhist.insert(blah);
  }

  info_to_stderr = 0;

  tty.grabwinsize();

  struct termios ti;
  struct termios oldti;
  
  tcgetattr(0, &oldti);
    
  cfmakeraw(&ti);
  tcsetattr(0, TCSADRAIN, &ti);

  setupterm(NULL, 0, NULL);
  signal(SIGWINCH, winch);

  main_loop();

  if (logfile)
    fclose(logfile);

  tcsetattr(0, TCSADRAIN, &oldti);
  printf("\n");
  return 0;
}
